// ComponentRegister.cpp: implementation of the ComponentRegister class.
//
//////////////////////////////////////////////////////////////////////

#include "ComponentRegister.h"
#include "ComponentVariant.h"
#include "ShiftedComponent.h"
#include "CompoundComponent.h"
#include "AcTree.h"
#include "AcTreeNode.h"
#include "Writer.h"
#include "Reader.h"
#include "Progress.h"

#include <assert.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ComponentRegister::ComponentRegister(const char * dataFileName)
{
	m_numComponents = 0;
	m_numVariants = 0;
	m_maxVariants = 1;

	BasicComponent * lastComp = NULL;

	char buffer[2000];

	// first we must load all components and variants, assign them IDs and determine their count

	Reader * rdr = new Reader(dataFileName);
	assert(rdr != NULL);

	if (!rdr->isOk())
	{
		sprintf(buffer, "Definition file cannot be opened: %s\n", dataFileName);
		goto error;
	}

	// first non-empty line should contain [Components]
	if (!rdr->readLine(buffer, sizeof(buffer)))
	{
		goto error_eof;
	}

	if (0 != strcmp(buffer, "[Components]"))
	{
		sprintf(buffer, "%s:%d: [Components] expected\n", dataFileName, rdr->getLineNo());
		goto error;
	}

	// Now read all components definitions

	{
		bool ok = true;

		do
		{
			if (!rdr->readLine(buffer, sizeof(buffer)))
			{
				goto error_eof;
			}
			// The line can contain next section header - [Pairs]
			if (0 == strcmp(buffer, "[Pairs]"))
			{
				ok = false;
			}
			else if ((buffer[0] == ' ') || (buffer[0] == '\t'))
			{
				// should be a variant definition
				if (lastComp == NULL)
				{
					sprintf(buffer, "%s:%d: Variant definition should be preceded by a component definition\n", 
						dataFileName, rdr->getLineNo());
					goto error;
				}

				ComponentVariant * var = parseVariant(buffer, lastComp);

				if (var == NULL)
				{
					sprintf(buffer, "%s:%d: Syntax error in variant definition\n",
						dataFileName, rdr->getLineNo());
					goto error;
				}

				m_variantQueue.push_back(var);
				++m_numVariants;
			}
			else
			{
				// should be a component definition

				lastComp = parseComponent(buffer);

				// all parameters collected
				m_componentQueue.push_back(lastComp);
				++m_numComponents;
			}
		} while (ok);
	}

	// Components and variants loaded

	// let's prepare the lists

	{
		m_componentList = new BasicComponent * [m_numComponents];
		assert(m_componentList != NULL);
		ComponentIterator it = m_componentQueue.begin();
		while (it != m_componentQueue.end())
		{
			m_componentList[(*it)->getId()] = (*it);
			if (m_maxVariants < (*it)->getQueueSize())
			{
				m_maxVariants = (*it)->getQueueSize();
			}
			++it;
		}
	}

	{
		m_variantList = new ComponentVariant * [m_numVariants];
		assert(m_variantList != NULL);
		VariantIterator it = m_variantQueue.begin();
		while (it != m_variantQueue.end())
		{
			m_variantList[(*it)->getId()] = (*it);
			++it;
		}
	}

	// and the variant matrix
	{
		ShiftedComponent sc1;
		ShiftedComponent sc2;

		m_variantMatrix = new bool * [m_numVariants];
		assert(m_variantMatrix != NULL);
		for (long i = 0; i < m_numVariants; i++)
		{
			sc1.init(getComponentById(getVariantById(i)->getComponentId()), 0, 0, TF);
			sc1.setVariant(getVariantById(i));

			m_variantMatrix[i] = new bool[m_numVariants];
			assert(m_variantMatrix[i] != NULL);
			for (long j = 0; j < m_numVariants; j++)
			{
				// let's check if variant j can follow after variant i
				ComponentVariant * varI = getVariantById(i);
				ComponentVariant * varJ = getVariantById(j);
				// if 'I' prohibits glider and 'J' releases it
				// or if 'I' requires glider and 'J' doesn't release it
				// mark this pair as invalid
				if ((varI->getGliderProhibited() && varJ->getGliderReleased())
					|| (varI->getGliderRequired() && !varJ->getGliderReleased()))
				{
					m_variantMatrix[i][j] = false;
				}
				else
				{
					sc2.init(getComponentById(getVariantById(j)->getComponentId()), sc1.getDx(), sc1.getDy(), sc1.getDirection());
					sc2.setVariant(getVariantById(j));
					// check if the variants can be put together
					m_variantMatrix[i][j] = sc1.variantCheck(&sc2);
				}
			}
		}	
	}

	// now let's extend the matrix by the second section

	{
		bool ok = true;

		do
		{
			if (!rdr->readLine(buffer, sizeof(buffer)))
			{
				goto error_eof;
			}
			// The line can contain next section header - [Sequences]
			if (0 == strcmp(buffer, "[Sequences]"))
			{
				ok = false;
			}
			else
			{
				char * cptr = buffer;
				char * cptr2;
				ComponentVariant * var1;
				ComponentVariant * var2;
				BasicComponent * comp1;
				BasicComponent * comp2;
				long i, j;

				// skip starting blanks

				while ((*cptr == ' ') || (*cptr == '\t'))
				{
					++cptr;
				}

				// possible syntaxes:
				// *<id>
				// <id>*
				// <id1>-<id2>
				// <id1>+<id2>

				if (*cptr == '*')
				{
					// let's guess it's "*<id>"
					++cptr;
					var1 = getVariantByName(cptr, true);
					if (var1 != NULL)
					{
						for (i = 0; i < m_numVariants; i++)
						{
							m_variantMatrix[i][var1->getId()] = false;
						}
					}
					else
					{
						comp1 = getComponentByName(cptr, false);
						if (comp1 != NULL)
						{
							for (j = 0; j < m_numVariants; j++)
							{
								if (getVariantById(j)->getComponentId() == comp1->getId())
								{
									for (i = 0; i < m_numVariants; i++)
									{
										m_variantMatrix[i][j] = false;
									}
								}
							}
						}
						else
						{
							sprintf(buffer, "%s:%d: Unrecognised pair\n", dataFileName, rdr->getLineNo());
							goto error;
						}
					}
				}
				else
				{
					cptr2 = strchr(cptr, '*');
					if (cptr2 != NULL)
					{
						// <id>*
						*cptr2 = '\0';
						var1 = getVariantByName(cptr, true);
						if (var1 != NULL)
						{
							for (i = 0; i < m_numVariants; i++)
							{
								m_variantMatrix[var1->getId()][i] = false;
							}
						}
						else
						{
							comp1 = getComponentByName(cptr, false);
							if (comp1 != NULL)
							{
								for (j = 0; j < m_numVariants; j++)
								{
									if (getVariantById(j)->getComponentId() == comp1->getId())
									{
										for (i = 0; i < m_numVariants; i++)
										{
											m_variantMatrix[j][i] = false;
										}
									}
								}
							}
							else
							{
								sprintf(buffer, "%s:%d: Unrecognised pair\n", dataFileName, rdr->getLineNo());
								goto error;
							}
						}
					}
					else
					{
						// <id1>+<id2> or <id1>-<id2>
						bool newst;

						cptr2 = strchr(cptr, '-');
						if (cptr2 != NULL)
						{
							newst = false;
						}
						else
						{
							newst = true;
							cptr2 = strchr(cptr, '+');
							if (cptr2 == NULL)
							{
								sprintf(buffer, "%s:%d: Unrecognised pair\n", dataFileName, rdr->getLineNo());
								goto error;
							}
						}
						*cptr2 = '\0';
						++cptr2;

						var1 = getVariantByName(cptr, true);
						if (var1 != NULL)
						{
							var2 = getVariantByName(cptr2, true);
							if (var2 != NULL)
							{
								m_variantMatrix[var1->getId()][var2->getId()] = newst;
							}
							else
							{
								comp2 = getComponentByName(cptr2, false);
								if (comp2 != NULL)
								{
									for (i = 0; i < m_numVariants; i++)
									{
										if (getVariantById(i)->getComponentId() == comp2->getId())
										{
											m_variantMatrix[var1->getId()][i] = newst;
										}
									}
								}
								else
								{
									sprintf(buffer, "%s:%d: Unrecognised pair\n", dataFileName, rdr->getLineNo());
									goto error;
								}
							}
						}
						else
						{
							comp1 = getComponentByName(cptr, false);
							if (comp1 != NULL)
							{
								var2 = getVariantByName(cptr2, true);
								if (var2 != NULL)
								{
									for (i = 0; i < m_numVariants; i++)
									{
										if (getVariantById(i)->getComponentId() == comp1->getId())
										{
											m_variantMatrix[i][var2->getId()] = newst;
										}
									}
								}
								else
								{
									comp2 = getComponentByName(cptr2, false);
									if (comp2 != NULL)
									{
										for (i = 0; i < m_numVariants; i++)
										{
											if (getVariantById(i)->getComponentId() == comp1->getId())
											{
												for (j = 0; j < m_numVariants; j++)
												{
													if (getVariantById(j)->getComponentId() == comp1->getId())
													{
														m_variantMatrix[i][j] = newst;
													}
												}
											}
										}
									}
									else
									{
										sprintf(buffer, "%s:%d: Unrecognised pair\n", dataFileName, rdr->getLineNo());
										goto error;
									}
								}
							}
							else
							{
								sprintf(buffer, "%s:%d: Unrecognised pair\n", dataFileName, rdr->getLineNo());
								goto error;
							}
						}

					}
				}
			}
		} while (ok);
	}

	// and now for the last part of the parser
	// let's collect all sequences and store them into the queue

	while (rdr->readLine(buffer, sizeof(buffer)))
	{
		CompoundComponent * cc = new CompoundComponent(AcTree::maxAcTreeDepth, m_maxVariants, AcTree::maxAcTreeDepth * m_maxVariants, "???");
		assert(cc != NULL);
		char * bptr = buffer;
		long n = 0;
		while (NULL != (bptr = strtok(bptr, " \t")))
		{
			++n;
			if (n > AcTree::maxAcTreeDepth)
			{
				sprintf(buffer, "%s:%d: Sequence too long\n", dataFileName, rdr->getLineNo());
				goto error;
			}
			BasicComponent * comp = getComponentByName(bptr, false);
			if (comp == NULL)
			{
				sprintf(buffer, "%s:%d: Invalid sequence\n", dataFileName, rdr->getLineNo());
				goto error;
			}
			cc->addComponent(comp);
			bptr = NULL;
		}
		if (n < 3)
		{
			sprintf(buffer, "%s:%d: Sequence too short\n", dataFileName, rdr->getLineNo());
			goto error;
		}

		m_sequenceList.push(cc);
	}

	// sequences will be inserted into the tree when it is built

	// now report

	char buf[80];
	sprintf(buf, "Loaded %d components, %d variants\n", m_numComponents, m_numVariants);
	Writer::print(buf);

	delete rdr;
	return;

error_eof:
	sprintf(buffer, "%s:%d: Unexpected end of file\n", dataFileName, rdr->getLineNo());
error:
	Writer::print(buffer);
	exit(1);
}

ComponentRegister::~ComponentRegister()
{
	delete m_acTree;

	long i;

	if (m_componentList != NULL)
	{
		for (i = 0; i < m_numComponents; i++)
		{
			delete m_componentList[i];
		}
		delete m_componentList;
		m_componentList = NULL;
	}
	
	if (m_variantList != NULL)
	{
		for (i = 0; i < m_numVariants; i++)
		{
			delete m_variantList[i];
		}
		delete m_variantList;
		m_variantList = NULL;
	}

	if (m_variantMatrix != NULL)
	{
		delete m_variantMatrix;
		m_variantMatrix = NULL;
	}
}

void ComponentRegister::prepare(long period, long compression, bool closed, bool useCache)
{
	long i, j;

	// prepare all components for specified period and compression
	// it will deactivate all variants that don't match
	for (i = 0; i < m_numComponents; i++)
	{
		m_componentList[i]->prepare(period, compression);
	}

	for (i = 0; i < m_numVariants; i++)
	{
		if (!m_variantList[i]->getActive())
		{
			// inactive variant can't precede or follow any other variant
			for (j = 0; j < m_numVariants; j++)
			{
				m_variantMatrix[i][j] = false;
				m_variantMatrix[j][i] = false;
			}	
		}
	}

	// remove all combinations that don't match

	// if closed loop, all variants must have active predecessor
	// and active successor
	// if not closed loop, all variants must have active predecessor
	// or active successor
	bool changed;
	bool predecessor, successor;

	do
	{
		changed = false;
		for (i = 0; i < m_numVariants; i++)
		{
			predecessor = false;
			successor = false;
			for (j = 0; j < m_numVariants; j++)
			{
				if (m_variantMatrix[i][j])
				{
					successor = true;
				}
				if (m_variantMatrix[j][i])
				{
					predecessor = true;
				}
			}

			if ((closed &&  (!predecessor || !successor))
				|| (!closed && !predecessor && !successor))
			{
				// variant can't be used - remove it from lists
				m_componentList[m_variantList[i]->getComponentId()]->deactivateVariant(m_variantList[i]);
				for (j = 0; j < m_numVariants; j++)
				{
					if (m_variantMatrix[i][j])
					{
						changed = true;
						m_variantMatrix[i][j] = false;
					}
					if (m_variantMatrix[j][i])
					{
						changed = true;
						m_variantMatrix[j][i] = false;
					}
				}	
			}
		}
	} while (changed);


	// calculate maximum number of variants a component can have
	m_maxVariants = 1;
	long n = 0;
	for (i = 0; i < m_numComponents; i++)
	{
		if (m_componentList[i]->getActive())
		{
			++n;

			long v = m_componentList[i]->getNumVariants();
			if (v > m_maxVariants)
			{
				m_maxVariants = v;
			}
		}
	}

	long v = 0;
	for (i = 0; i < m_numVariants; i++)
	{
		ComponentVariant * vptr = getVariantById(i);
		if (vptr->getActive())
		{
			if (useCache)
			{
				vptr->setUseCache();
			}
			++v;
		}
	}

	// component preparations complete
	// write a report

	char buf[100];
	sprintf(buf, "Prepared %d components, %d variants\n", n, v);
	Writer::print(buf);
}

bool ComponentRegister::canFollow(ComponentVariant *var1, ComponentVariant *var2)
{
	return m_variantMatrix[var1->getId()][var2->getId()];
}

bool ComponentRegister::canFollow(ComponentVariant *var, BasicComponent *com)
{
	long n = com->getNumVariants();
	long id = var->getId();
	for (long i = 0; i < n; i++)
	{
		if (m_variantMatrix[id][com->getActiveVariant(i)->getId()])
		{
			return true;
		}
	}
	return false;
}

bool ComponentRegister::canFollow(BasicComponent *com, ComponentVariant *var)
{
	long n = com->getNumVariants();
	long id = var->getId();
	for (long i = 0; i < n; i++)
	{
		if (m_variantMatrix[com->getActiveVariant(i)->getId()][id])
		{
			return true;
		}
	}
	return false;
}

bool ComponentRegister::canFollow(BasicComponent *com1, BasicComponent *com2)
{
	long n1 = com1->getNumVariants();
	long n2 = com2->getNumVariants();

	long i;
	long j;

	for (i = 0; i < n1; i++)
	{
		long id1 = com1->getActiveVariant(i)->getId();
		for (j = 0; j < n2; j++)
		{
			if (m_variantMatrix[id1][com2->getActiveVariant(j)->getId()])
			{
				return true;
			}
		}
	}
	return false;
}

void ComponentRegister::buildTree(long treeSize)
{
	Writer::print("Building Aho/Corasick tree...\n");

	// the tree

	m_acTree = new AcTree(m_numComponents);
	assert(m_acTree != NULL);

	long i;

	CompoundComponent * cc = new CompoundComponent(AcTree::maxAcTreeDepth, getMaxVariants(), AcTree::maxAcTreeDepth * getMaxVariants(), "???");
	assert(cc != NULL);

	// add all inactive components

	for(i = 0; i < m_numComponents; i++)
	{
		if (!m_componentList[i]->getActive())
		{
			cc->addComponent(m_componentList[i]);
			m_acTree->addComponent(cc);
			cc->removeLastComponent();
		}
	}

	// add all manually entered sequences
	while (m_sequenceList.size() > 0)
	{
		m_acTree->addComponent(m_sequenceList.top());
		delete m_sequenceList.top();
		m_sequenceList.pop();
	}

	m_finishedTreeLevel = 1;

	// generate other sequences

	treeSize = treeSize / (sizeof(AcTreeNode) + m_numComponents * sizeof(AcTreeNode *));
	Progress::activate(treeSize);
	i = 2;
	while ((i <= AcTree::maxAcTreeDepth) && (m_acTree->getNumNodes() <= treeSize))
	{
		if (buildTreeLevel(cc, i, treeSize))
		{
			m_finishedTreeLevel = i;
		}
		++i;
	}

	// finish the structure

	m_acTree->finish();

	Progress::cleanup();

	{
		char buf[100];
		sprintf(buf, "Tree built: %d nodes, %d sequences, complete to depth %d\n", 
			m_acTree->getNumNodes(), 
			m_acTree->getNumSequences(),
			m_finishedTreeLevel);
		Writer::print(buf);
	}

	delete cc;
}

AcTreeNode * ComponentRegister::getAcTreeRoot()
{
	return m_acTree->getRoot();
}

bool ComponentRegister::buildTreeLevel(CompoundComponent * cc, long l, long maxNodes)
{
	long i;
	for (i = 0; i < m_numComponents; i++)
	{
		if (m_componentList[i]->getActive())
		{
			cc->addComponent(m_componentList[i]);
			if (m_acTree->checkComponent(cc))
			{
				if (l > 1)
				{
					if (!buildTreeLevel(cc, l - 1, maxNodes))
					{
						cc->removeLastComponent();
						return false;
					}
				}
				else
				{
					if (0 != cc->check(false))
					{
						m_acTree->addComponent(cc);

						Progress::setStep(m_acTree->getNumNodes());

						if (m_acTree->getNumNodes() > maxNodes)
						{
							cc->removeLastComponent();
							return false;
						}

					}
				}
			}
			cc->removeLastComponent();
		}
	}
	return true;
}

BasicComponent * ComponentRegister::parseComponent(char *buffer)
{
	char * name = strtok(buffer, " \t");
	if (name == NULL)
	{
		Writer::print("Missing component name\n");
		return NULL;
	}

	if (getComponentByName(name, true) != NULL)
	{
		char buf[80];
		sprintf(buf, "Duplicate component name: %s\n", name);
		Writer::print(buf);
		return NULL;
	}

	char * token = strtok(NULL, " \t");
	if (token == NULL)
	{
		Writer::print("Missing direction\n");
		return NULL;
	}

	Direction dir;

	if (0 == strcmp(token, "F"))
	{
		dir = TF;
	}
	else if (0 == strcmp(token, "Fx"))
	{
		dir = TFX;
	}
	else if (0 == strcmp(token, "R"))
	{
		dir = TR;
	}
	else if (0 == strcmp(token, "Rx"))
	{
		dir = TRX;
	}
	else if (0 == strcmp(token, "L"))
	{
		dir = TL;
	}
	else if (0 == strcmp(token, "Lx"))
	{
		dir = TLX;
	}
	else if (0 == strcmp(token, "B"))
	{
		dir = TB;
	}
	else if (0 == strcmp(token, "Bx"))
	{
		dir = TBX;
	}
	else
	{
		char buf[80];
		sprintf(buf, "Unrecognized direction: %s\n", token);
		Writer::print(buf);
		return NULL;
	}

	long dt, dx, dy;

	if (!parseInteger(1, 1000000, dt))
	{
		Writer::print("Invalid DT\n");
		return NULL;
	}

	if (!parseInteger(-1000000, 1000000, dx))
	{
		Writer::print("Invalid DX\n");
		return NULL;
	}

	if (!parseInteger(-1000000, 1000000, dy))
	{
		Writer::print("Invalid DY\n");
		return NULL;
	}

	BasicComponent * bc = new BasicComponent(this, m_numComponents, name, dir, dx, dy, dt);
	assert(bc != NULL);

	return bc;
}

bool ComponentRegister::parseInteger(long minValue, long maxValue, long &value)
{
	char * token = strtok(NULL, " \t");
	char * ep;

	if (token == NULL)
	{
		Writer::print("Missing integer value\n");
		return false;
	}
	
	value = strtol(token, &ep, 10);
	if (*ep != '\0')
	{
		char buf[80];
		sprintf(buf, "Invalid integer: %s\n", token);
		Writer::print(buf);
		return false;
	}

	if (minValue > value)
	{
		char buf[80];
		sprintf(buf, "Integer value too small (%d) - limit is %d\n", value, minValue);
		Writer::print(buf);
		return false;
	}

	if (maxValue < value)
	{
		char buf[80];
		sprintf(buf, "Integer value too great (%d) - limit is %d\n", value, maxValue);
		Writer::print(buf);
		return false;
	}

	return true;
}

ComponentVariant * ComponentRegister::parseVariant(char *buffer, BasicComponent *lastComp)
{
	char * name = strtok(buffer, " \t");
	if (name == NULL)
	{
		Writer::print("Missing variant name\n");
		return NULL;
	}

	if (getVariantByName(name, true) != NULL)
	{
		char buf[80];
		sprintf(buf, "Duplicate variant name: %s\n", name);
		Writer::print(buf);
		return NULL;
	}

	long period, maxc, minc;

	if (!parseInteger(1, 1000000, period))
	{
		Writer::print("Invalid period\n");
		return NULL;
	}

	if (!parseInteger(1, 1000000, maxc))
	{
		Writer::print("Invalid maximum compression\n");
		return NULL;
	}

	if (!parseInteger(0, 1000000, minc) || ((minc != 0) && (minc < maxc)))
	{
		Writer::print("Invalid minimum compression\n");
		return NULL;
	}

	char * token = strtok(NULL, " \t");
	if (*token != 'G')
	{
		Writer::print("Invalid variant type\n");
		return NULL;
	}
	++token;
	// follows either R or S
	bool grel = false;
	bool greq = false;
	bool gproh = false;

	if (*token == 'R')
	{
		grel = true;
	}
	else if (*token != 'S')
	{
		Writer::print("Invalid variant type\n");
		return NULL;
	}
	++token;
	// the string may end
	// or 'U' or 'X' may follow
	if (*token == 'U')
	{
		greq = true;
		++token;
	}
	else if (*token == 'X')
	{
		gproh = true;
		++token;
	}
	// now the token should end
	if (*token != '\0')
	{
		Writer::print("Invalid variant type\n");
		return NULL;
	}

	CellMap * image = parseCellMap(true);
	if (image == NULL)
	{
		Writer::print("Invalid variant image\n");
		return NULL;
	}

	CellMap * usage = parseCellMap(false);
	if (usage == NULL)
	{
		image->delRef();
		Writer::print("Invalid variant usage\n");
		return NULL;
	}

	ComponentVariant * var = new ComponentVariant(
		name, m_numVariants, lastComp->getId(), period, 
		maxc, minc, grel, greq, gproh, image, usage);
	assert(var != NULL);
	lastComp->addVariant(var);

	return var;
}

CellMap * ComponentRegister::parseCellMap(bool img)
{
	char * token = strtok(NULL, " \t");
	if (token == NULL)
	{
		Writer::print("Missing cell map\n");
		return NULL;
	}

	if (*token == '=')
	{
		// use another component
		++token;

		ComponentVariant * var = getVariantByName(token, false);
		if (var == NULL)
		{
			char buf[80];
			sprintf(buf, "Invalid variant name: %s\n", token);
			Writer::print(buf);
			return NULL;
		}

		CellMap * cm;

		if (img)
		{
			cm = var->getImageMap();
		}
		else
		{
			cm = var->getUsageMap();
		}
		cm->addRef();
		return cm;
	}
	else
	{
		// use a file
		Reader * rdr = new Reader(token);
		assert(rdr != NULL);
		if (!rdr->isOk())
		{
			char buf[600];
			sprintf(buf, "Invalid image filename: %s\n", token);
			Writer::print(buf);
			delete rdr;
			return NULL;
		}

		char rlebuf[10000];
		char * rlepos = rlebuf;
		long rem = sizeof(rlebuf) - 1;
		bool bang = false;

		while (!bang && (rem > 0) && rdr->readLine(rlepos, rem))
		{
			if ((rlepos != rlebuf)
				|| (((*rlepos >= '0') && (*rlepos <= '9'))
					|| (*rlepos == 'b')
					|| (*rlepos == 'o')))
			{
				bang = (NULL != strchr(rlepos, '!'));
				long r = strlen(rlepos);
				rlepos += r;
				rem -= r;
				while ((rlepos > rlebuf) && (*(rlepos - 1) < ' '))
				{
					--rlepos;
					*rlepos = '\0';
					++rem;
				}
			}
		}

		if ((!bang) || (rlepos == rlebuf) || (rem <= 0))
		{
			char buf[600];
			sprintf(buf, "Invalid image file: %s\n", token);
			Writer::print(buf);
			delete rdr;
			return NULL;
		}

		long dx, dy;

		if (!parseInteger(-1000000, 1000000, dx))
		{
			Writer::print("Invalid cell map DX\n");
			return NULL;
		}

		if (!parseInteger(-1000000, 1000000, dy))
		{
			Writer::print("Invalid cell map DY\n");
			return NULL;
		}

		delete rdr;

		CellMap * cm = new CellMap(dx, dy, rlebuf);
		assert(cm != NULL);

		return cm;
	}
}

BasicComponent * ComponentRegister::getComponentByName(const char *name, bool quiet)
{
	ComponentIterator it;
	it = m_componentQueue.begin();
	while (it != m_componentQueue.end())
	{
		if (0 == strcmp(name, (*it)->getName()))
		{
			return (*it);
		}
		++it;
	}
	if (!quiet)
	{
		char buf[600];
		sprintf(buf, "Unknown component name: %s\n", name);
		Writer::print(buf);
	}
	return NULL;
}

ComponentVariant * ComponentRegister::getVariantByName(const char *name, bool quiet)
{
	VariantIterator it;
	it = m_variantQueue.begin();
	while (it != m_variantQueue.end())
	{
		if (0 == strcmp(name, (*it)->getName()))
		{
			return (*it);
		}
		++it;
	}
	if (!quiet)
	{
		char buf[600];
		sprintf(buf, "Unknown variant name: %s\n", name);
		Writer::print(buf);
	}
	return NULL;
}

CompoundComponent * ComponentRegister::getInvalidSequence()
{
	if (m_sequenceList.size() <= 0)
	{
		return NULL;
	}
	else
	{
		CompoundComponent * comp = m_sequenceList.top();
		m_sequenceList.pop();
		return comp;
	}
}
