// ComponentVariant.cpp: implementation of the ComponentVariant class.
//
//////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <string.h>
#include "ComponentVariant.h"
#include "Component.h"
#include <assert.h>
#include <stdio.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ComponentVariant::ComponentVariant(
	char * name,
	long id, 
	long componentId,
	long period,
	long maxCompression,
	long minCompression,
	bool gliderReleased,
	bool gliderRequired,
	bool gliderProhibited,
	CellMap * image,
	CellMap * usage)
{
	m_active = true;

	m_name = new char[strlen(name) + 1];
	assert(m_name != NULL);

	strcpy(m_name, name);

	m_id = id;
	m_componentId = componentId;

	m_gliderReleased = gliderReleased;
	m_gliderRequired = gliderRequired;
	m_gliderProhibited = gliderProhibited;

	m_period = period;
	m_maxCompression = maxCompression;
	m_minCompression = minCompression;

	m_image = image; // expect image and usage to have reference count increased already
	m_usage = usage;

	m_useCache = false;

	long loX, hiX, loY, hiY;
	m_image->getSize(loX, hiX, loY, hiY);
	for (long i = 0; i < 8; i++)
	{
		Component::localToGlobal(
			0, 0, (Direction) i, 
			loX, hiX, loY, hiY, 
			&m_xMin[i], &m_xMax[i], &m_yMin[i], &m_yMax[i]);

		m_checkCache[i] = NULL;
	}
}

ComponentVariant::~ComponentVariant()
{
	m_image->delRef();
	m_image = NULL;
	m_usage->delRef();
	m_usage = NULL;
	delete [] m_name;
	delete [] m_checkCache[0];
	delete [] m_checkCache[1];
	delete [] m_checkCache[2];
	delete [] m_checkCache[3];
	delete [] m_checkCache[4];
	delete [] m_checkCache[5];
	delete [] m_checkCache[6];
	delete [] m_checkCache[7];
}

bool ComponentVariant::query(long period, long compression)
{
	m_active = ((period == 0) || (period % m_period == 0))
		&& ((compression == 0)
			|| ((compression != 0)
				&& (compression >= m_maxCompression)
				&& ((m_minCompression == 0)
					|| (m_minCompression >= compression))));
	return m_active;
}

void ComponentVariant::deactivate()
{
	m_active = false;
}

// the ugly constructions with endless loops etc are due to processor cache hits
// the 'normal' way (i.e. check cache hit) should be as straight and as short as possible

bool ComponentVariant::check(
	long x, 
	long y, 
	Direction d, 
	ComponentVariant * v)
{
	if (m_useCache)
	{
		CheckCacheRecord * ccr = m_checkCache[d] + v->getId();
		do
		{
			if (ccr->array != NULL)
			{
				long qx = x - ccr->xMin;
				if ((qx < 0)
					|| (qx >= ccr->width))
				{
					return true;
				}
				long qy = y - ccr->yMin;
				if ((qy < 0)
					|| (qy >= ccr->height))
				{
					return true;
				}

				char * cca = ccr->array + (qy * ccr->width + qx);

				if (*cca != 0)
				{
					return (*cca == 1);
				}

				// the point wasn't checked yet
				// so let's do the check

				
				long vx1, vx2, vy1, vy2;
				v->getSize(d, vx1, vx2, vy1, vy2);
				vx1 += x;
				vx2 += x;
				vy1 += y;
				vy2 += y;

				if (m_xMin[0] > vx1)
				{
					vx1 = m_xMin[0];
				}
				if (m_xMax[0] < vx2)
				{
					vx2 = m_xMax[0];
				}
				if (m_yMin[0] > vy1)
				{
					vy1 = m_yMin[0];
				}
				if (m_yMax[0] < vy2)
				{
					vy2 = m_yMax[0];
				}

				if ((vx1 <= vx2) && (vy1 <= vy2))
				{
					long ry = vy1;
					do
					{
						long rx = vx1;
						do
						{
							if (m_usage->getCell(rx, ry))
							{
								long sx, sy;
								Component::globalToLocal(x, y, d, rx, ry, &sx, &sy);
								if (v->getUsage(sx, sy))
								{
									// variants collide
									*cca = 2;
									return false;
								}
							}
							++rx;
						} while (rx <= vx2);
						++ry;
					} while (ry <= vy2);
				}

				// variants don't collide
				*cca = 1;
				return true;
			}

			// the array for this variant pair isn't initialized
			// get the variant's bounding rectangle
			long xMin, xMax, yMin, yMax;
			v->getSize(d, xMin, xMax, yMin, yMax);
			ccr->xMin = m_xMin[0] - xMax;
			ccr->width = (m_xMax[0] - xMin) - ccr->xMin + 1;
			ccr->yMin = m_yMin[0] - yMax;
			ccr->height = (m_yMax[0] - yMin) - ccr->yMin + 1;

			// allocate the array
			long n = ccr->width * ccr->height;
			ccr->array = new char[n];
			assert(ccr->array != NULL);

			// initialize the array contents
			char * c = ccr->array;
			do
			{
				*c = 0;
				++c;
				--n;
			} while (n != 0);

		} while (true);
	}
	else
	{
		// memory saving branch
		// no cache use, just check both variants
		long vx1, vx2, vy1, vy2;
		v->getSize(d, vx1, vx2, vy1, vy2);
		vx1 += x;
		vx2 += x;
		vy1 += y;
		vy2 += y;

		if (m_xMin[0] > vx1)
		{
			vx1 = m_xMin[0];
		}
		if (m_xMax[0] < vx2)
		{
			vx2 = m_xMax[0];
		}
		if (m_yMin[0] > vy1)
		{
			vy1 = m_yMin[0];
		}
		if (m_yMax[0] < vy2)
		{
			vy2 = m_yMax[0];
		}

		if ((vx1 <= vx2) && (vy1 <= vy2))
		{
			long ry = vy1;
			do
			{
				long rx = vx1;
				do
				{
					if (m_usage->getCell(rx, ry))
					{
						long sx, sy;
						Component::globalToLocal(x, y, d, rx, ry, &sx, &sy);
						if (v->getUsage(sx, sy))
						{
							// variants collide
							return false;
						}
					}
					++rx;
				} while (rx <= vx2);
				++ry;
			} while (ry <= vy2);
		}
		return true;
	}
}



void ComponentVariant::setUseCache()
{
	m_useCache = true;

	for (long i = 0; i < 8; i++)
	{
		long n = m_id + 1;

		m_checkCache[i] = new CheckCacheRecord[n];
		assert(m_checkCache[i] != NULL);

		do
		{
			--n;
			m_checkCache[i][n].array = NULL;
		} while (n != 0);	
	}
}
